# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
S3 buckets and folders creation, deletion and configuration."""

from typing import Dict, List
from botocore.exceptions import ClientError
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception
import scar.logger as logger


class S3Client(BotoClient):
    """A low-level client representing Amazon Simple Storage Service (S3Client).
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/services/s3.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 's3'
    _DEFAULT_ACL = 'private'

    @exception(logger)
    def create_bucket(self, bucket_name: str) -> Dict:
        """Creates a new S3 bucket."""
        return self.client.create_bucket(ACL=self._DEFAULT_ACL, Bucket=bucket_name)

    @exception(logger)
    def find_bucket(self, bucket_name: str) -> bool:
        """Checks bucket existence."""
        try:
            # If this call works the bucket exists
            self.client.get_bucket_location(Bucket=bucket_name)
            return True
        except ClientError as cerr:
            # Function not found
            if cerr.response['Error']['Code'] == 'NoSuchBucket':
                return False
            raise cerr

    @exception(logger)
    def put_notification_configuration(self, bucket_name: str, notification: Dict) -> Dict:
        """Enables notifications of specified events for a bucket."""
        kwargs = {"Bucket": bucket_name,
                  "NotificationConfiguration": notification}
        return self.client.put_bucket_notification_configuration(**kwargs)

    @exception(logger)
    def get_notification_configuration(self, bucket_name: str) -> Dict:
        """Returns the notification configuration of a bucket."""
        return self.client.get_bucket_notification_configuration(Bucket=bucket_name)

    @exception(logger)
    def upload_file(self, **kwargs: Dict) -> Dict:
        """Adds an object to a bucket."""
        return self.client.put_object(**kwargs)

    @exception(logger)
    def download_file(self, **kwargs: Dict) -> Dict:
        """Download an object from S3 to a file-like object."""
        return self.client.download_fileobj(**kwargs)

    @exception(logger)
    def is_folder(self, bucket: str, folder: str) -> bool:
        """Checks if a file with the key specified exists."""
        try:
            kwargs = {'Bucket' : bucket,
                      'Key' : folder if folder.endswith('/') else folder + '/'}
            # If this call works the folder exist
            self.client.get_object(**kwargs)
            return True
        except ClientError as cerr:
            # Folder not found
            if cerr.response['Error']['Code'] == 'NoSuchKey':
                return False
            raise cerr

    @exception(logger)
    def list_files(self, **kwargs: Dict) -> List:
        """Returns the keys of all the objects in a bucket.
        Excludes the S3 'folders', i.e. files with key ending in '/'."""
        file_list = []
        response = self.client.list_objects_v2(**kwargs)
        if 'Contents' in response:
            # Don't append S3 'folders'
            file_list.extend([file['Key'] for file in response['Contents']
                              if not file['Key'].endswith('/')])
        # Retrieve all the remaining file keys recursively
        if response['IsTruncated']:
            kwargs['ContinuationToken'] = response['NextContinuationToken']
            file_list.extend(self.list_files(**kwargs))
        return file_list



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Tuple, Dict, List
from scar.providers.aws import GenericClient
import scar.exceptions as excp
import scar.logger as logger
from scar.utils import FileUtils


def get_bucket_and_folders(storage_path: str) -> Tuple:
    output_bucket = storage_path
    output_folders = ""
    output_path = storage_path.split("/", 1)
    if len(output_path) > 1:
        # There are folders defined
        output_bucket = output_path[0]
        output_folders = output_path[1]
    return (output_bucket, output_folders)


class S3(GenericClient):

    def __init__(self, resources_info):
        super().__init__(resources_info.get('s3'))
        self.resources_info = resources_info

    @excp.exception(logger)
    def create_bucket(self, bucket_name) -> None:
        if not self.client.find_bucket(bucket_name):
            self.client.create_bucket(bucket_name)

    @excp.exception(logger)
    def add_bucket_folder(self, bucket: str, folders: str) -> None:
        if not self.client.is_folder(bucket, folders):
            self.upload_file(bucket, folder_name=folders)

    def create_bucket_and_folders(self, storage_path: str) -> Tuple:
        bucket, folders = get_bucket_and_folders(storage_path)
        self.create_bucket(bucket)
        if folders:
            self.add_bucket_folder(bucket, folders)
        return bucket, folders

    def set_input_bucket_notification(self, bucket_name: str, folders: str) -> None:
        # First check that the function doesn't have other configurations
        bucket_conf = self.client.get_notification_configuration(bucket_name)
        trigger_conf = self.get_trigger_configuration(folders)
        lambda_conf = [trigger_conf]
        if "LambdaFunctionConfigurations" in bucket_conf:
            lambda_conf = bucket_conf["LambdaFunctionConfigurations"]
            lambda_conf.append(trigger_conf)
        notification = {"LambdaFunctionConfigurations": lambda_conf}
        self.client.put_notification_configuration(bucket_name, notification)

    def delete_bucket_notification(self, bucket_name):
        bucket_conf = self.client.get_notification_configuration(bucket_name)
        if bucket_conf and "LambdaFunctionConfigurations" in bucket_conf:
            lambda_conf = bucket_conf["LambdaFunctionConfigurations"]
            filter_conf = [x for x in lambda_conf if x['LambdaFunctionArn'] != self.resources_info.get('lambda').get('arn')]
            notification = {"LambdaFunctionConfigurations": filter_conf}
            self.client.put_notification_configuration(bucket_name, notification)
            logger.info("Bucket notifications successfully deleted.")

    def get_trigger_configuration(self, folders: str) -> Dict:
        conf = {"LambdaFunctionArn": self.resources_info.get('lambda').get('arn'),
                "Events": ["s3:ObjectCreated:*"]}
        if folders != '':
            conf['Filter'] = {"Key": {"FilterRules": [{"Name": "prefix", "Value": f'{folders}/'}]}}
        return conf

    def get_file_key(self, folder_name=None, file_path=None, file_key=None):
        if file_key:
            return file_key
        file_key = ''
        if file_path:
            file_key = os.path.basename(file_path)
            if folder_name:
                file_key = FileUtils.join_paths(folder_name, file_key)
        elif folder_name:
            file_key = folder_name if folder_name.endswith('/') else '{0}/'.format(folder_name)
        return file_key

    @excp.exception(logger)
    def upload_file(self, bucket: str, folder_name: str=None, file_path: str=None, file_key: str=None) -> None:
        kwargs = {'Bucket': bucket}
        kwargs['Key'] = self.get_file_key(folder_name, file_path, file_key)
        if file_path:
            try:
                kwargs['Body'] = FileUtils.read_file(file_path, 'rb')
            except FileNotFoundError:
                raise excp.UploadFileNotFoundError(file_path=file_path)
        if folder_name and not file_path:
            kwargs['ContentType'] = 'application/x-directory'
            logger.info(f"Folder '{kwargs['Key']}' created in bucket '{kwargs['Bucket']}'.")
        else:
            logger.info(f"Uploading file '{file_path}' to bucket '{kwargs['Bucket']}' with key '{kwargs['Key']}'.")
        self.client.upload_file(**kwargs)

    @excp.exception(logger)
    def get_bucket_file_list(self, storage: Dict=None):
        files = []
        if storage:
            files = self._list_storage_files(storage)
        else:
            for storage_info in self.resources_info.get('lambda').get('input'):
                if storage_info.get('storage_provider') == 's3':
                    files.extend(self._list_storage_files(storage_info))
        return files

    def _list_storage_files(self, storage: Dict) -> List:
        files = []
        bucket_name, folder_path = get_bucket_and_folders(storage.get('path'))
        if self.client.find_bucket(bucket_name):
            kwargs = {"Bucket" : bucket_name}
            if folder_path:
                kwargs["Prefix"] = folder_path
            files = (self.client.list_files(**kwargs))
        else:
            raise excp.BucketNotFoundError(bucket_name=bucket_name)
        return files

    def get_s3_event(self, bucket_name, file_key):
        event = self.resources_info.get("s3").get("event")
        event['Records'][0]['s3']['bucket']['name'] = bucket_name
        event['Records'][0]['s3']['bucket']['arn'] = event['Records'][0]['s3']['bucket']['arn'].format(bucket_name=bucket_name)
        event['Records'][0]['s3']['object']['key'] = file_key
        return event

    def get_s3_event_list(self, bucket_name, file_keys):
        return [self.get_s3_event(bucket_name, file_key) for file_key in file_keys]

    def download_file(self, bucket_name, file_key, file_path):
        kwargs = {'Bucket' : bucket_name, 'Key' : file_key}
        logger.info(f"Downloading file '{file_key}' from bucket '{bucket_name}' in path '{file_path}'.")
        with open(file_path, 'wb') as file:
            kwargs['Fileobj'] = file
            self.client.download_file(**kwargs)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import io
from typing import Dict
from multiprocessing.pool import ThreadPool
from zipfile import ZipFile, BadZipfile
import yaml
from botocore.exceptions import ClientError
from scar.http.request import call_http_endpoint, get_file
from scar.providers.aws import GenericClient
from scar.providers.aws.functioncode import FunctionPackager
from scar.providers.aws.lambdalayers import LambdaLayers
from scar.providers.aws.s3 import S3
from scar.providers.aws.validators import AWSValidator
import scar.exceptions as excp
import scar.logger as logger
from scar.utils import DataTypesUtils, FileUtils, StrUtils, SupervisorUtils
from scar.parser.cfgfile import ConfigFileParser


MAX_CONCURRENT_INVOCATIONS = 500
ASYNCHRONOUS_CALL = {"invocation_type": "Event",
                     "log_type": "None",
                     "asynchronous": "True"}
REQUEST_RESPONSE_CALL = {"invocation_type": "RequestResponse",
                         "log_type": "Tail",
                         "asynchronous": "False"}


class Lambda(GenericClient):

    def __init__(self, resources_info: Dict) -> None:
        super().__init__(resources_info.get('lambda', {}))
        self.resources_info = resources_info
        self.function = resources_info.get('lambda', {})
        self.supervisor_version = resources_info.get('lambda').get('supervisor').get('version')

    def _get_creations_args(self, zip_payload_path: str, supervisor_zip_path: str) -> Dict:
        return {'FunctionName': self.function.get('name'),
                'Runtime': self.function.get('runtime'),
                'Role': self.resources_info.get('iam').get('role'),
                'Handler': self.function.get('handler'),
                'Code': self._get_function_code(zip_payload_path, supervisor_zip_path),
                'Environment': self.function.get('environment'),
                'Description': self.function.get('description'),
                'Timeout':  self.function.get('timeout'),
                'MemorySize': self.function.get('memory'),
                'Tags': self.function.get('tags'),
                'Layers': self.function.get('layers')}

    def is_asynchronous(self):
        return self.function.get('asynchronous', False)

    def get_access_key(self) -> str:
        """Returns the access key belonging to the boto_profile used."""
        return self.client.get_access_key()

    @excp.exception(logger)
    def create_function(self):
        # Create tmp folders
        supervisor_path = FileUtils.create_tmp_dir()
        tmp_folder = FileUtils.create_tmp_dir()
        # Download supervisor
        supervisor_zip_path = SupervisorUtils.download_supervisor(
            self.supervisor_version,
            supervisor_path.name
        )
        # Manage supervisor layer
        self._manage_supervisor_layer(supervisor_zip_path)
        # Create function
        zip_payload_path = FileUtils.join_paths(tmp_folder.name, 'function.zip')
        self._set_image_id()
        creation_args = self._get_creations_args(zip_payload_path, supervisor_zip_path)
        response = self.client.create_function(**creation_args)
        if response and "FunctionArn" in response:
            self.function['arn'] = response.get('FunctionArn', "")
        return response

    def _set_image_id(self):
        image = self.function.get('container').get('image')
        if image:
            self.function['environment']['Variables']['IMAGE_ID'] = image

    def _manage_supervisor_layer(self, supervisor_zip_path: str) -> None:
        layers_client = LambdaLayers(self.resources_info, self.client, supervisor_zip_path)
        self.function.get('layers', []).append(layers_client.get_supervisor_layer_arn())

    @excp.exception(logger)
    def _get_function_code(self, zip_payload_path: str, supervisor_zip_path: str) -> Dict:
        '''Zip all the files and folders needed.'''
        code = {}
        FunctionPackager(self.resources_info, supervisor_zip_path).create_zip(zip_payload_path)
        if self.function.get('deployment').get('bucket', False):
            file_key = f"lambda/{self.function.get('name')}.zip"
            s3_client = S3(self.resources_info)
            s3_client.create_bucket(self.function.get('deployment').get('bucket'))
            s3_client.upload_file(bucket=self.function.get('deployment').get('bucket'),
                                  file_path=zip_payload_path,
                                  file_key=file_key)
            code = {"S3Bucket": self.function.get('deployment').get('bucket'),
                    "S3Key": file_key}
        else:
            code = {"ZipFile": FileUtils.read_file(zip_payload_path, mode="rb")}
        return code

    def delete_function(self):
        return self.client.delete_function(self.resources_info.get('lambda').get('name'))

    def link_function_and_bucket(self, bucket_name: str) -> None:
        kwargs = {'FunctionName': self.function.get('name'),
                  'Principal': "s3.amazonaws.com",
                  'SourceArn': f'arn:aws:s3:::{bucket_name}'}
        self.client.add_invocation_permission(**kwargs)

    def preheat_function(self):
        logger.info("Preheating function")
        self._set_request_response_call_parameters()
        self.launch_lambda_instance()
        logger.info("Preheating successful")

    def _launch_async_event(self, s3_event):
        self.set_asynchronous_call_parameters()
        return self._launch_s3_event(s3_event)

    def launch_request_response_event(self, s3_event):
        self._set_request_response_call_parameters()
        return self._launch_s3_event(s3_event)

    def _launch_s3_event(self, s3_event):
        self.function['payload'] = s3_event
        logger.info(f"Sending event for file '{s3_event['Records'][0]['s3']['object']['key']}'")
        return self.launch_lambda_instance()

    def process_asynchronous_lambda_invocations(self, s3_event_list):
        if (len(s3_event_list) > MAX_CONCURRENT_INVOCATIONS):
            for s3_file_chunk in DataTypesUtils.divide_list_in_chunks(s3_event_list, MAX_CONCURRENT_INVOCATIONS):
                self._launch_concurrent_lambda_invocations(s3_file_chunk)
        else:
            self._launch_concurrent_lambda_invocations(s3_event_list)

    def _launch_concurrent_lambda_invocations(self, s3_event_list):
        pool = ThreadPool(processes=len(s3_event_list))
        pool.map(lambda s3_event: self._launch_async_event(s3_event), s3_event_list)
        pool.close()

    def launch_lambda_instance(self):
        if self.is_asynchronous():
            self.set_asynchronous_call_parameters()
        response = self._invoke_lambda_function()
        response_args = {'Response' : response,
                         'FunctionName' : self.function.get('name'),
                         'IsAsynchronous' : self.function.get('asynchronous')}
        return response_args

    def _get_invocation_payload(self):
        # Default payload
        payload = self.function.get('payload', {})
        if not payload:
            # Check for defined run script
            if self.function.get("run_script", False):
                script_path = self.function.get("run_script")
                # We first code to base64 in bytes and then decode those bytes to allow the json lib to parse the data
                # https://stackoverflow.com/questions/37225035/serialize-in-json-a-base64-encoded-data#37239382
                payload = {"script": StrUtils.bytes_to_base64str(FileUtils.read_file(script_path, 'rb'))}
            # Check for defined commands
            # This overrides any other function payload
            if self.function.get("c_args", False):
                payload = {"cmd_args": json.dumps(self.function.get("c_args"))}
        return json.dumps(payload)

    def _invoke_lambda_function(self):
        invoke_args = {'FunctionName':  self.function.get('name'),
                       'InvocationType':  self.function.get('invocation_type'),
                       'LogType':  self.function.get('log_type'),
                       'Payload': self._get_invocation_payload()}
        return self.client.invoke_function(**invoke_args)

    def set_asynchronous_call_parameters(self):
        self.function.update(ASYNCHRONOUS_CALL)

    def _set_request_response_call_parameters(self):
        self.function.update(REQUEST_RESPONSE_CALL)

    def _get_function_environment_variables(self):
        return self.get_function_configuration()['Environment']

    def merge_aws_and_local_configuration(self, aws_conf: Dict) -> Dict:
        result = ConfigFileParser().get_properties().get('aws')
        result['lambda']['name'] = aws_conf['FunctionName']
        result['lambda']['arn'] = aws_conf['FunctionArn']
        result['lambda']['timeout'] = aws_conf['Timeout']
        result['lambda']['memory'] = aws_conf['MemorySize']
        result['lambda']['environment']['Variables'] = aws_conf['Environment']['Variables'].copy()
        result['lambda']['layers'] = aws_conf['Layers'].copy()
        result['lambda']['supervisor']['version'] = aws_conf['SupervisorVersion']
        return result

    def get_all_functions(self, arn_list):
        try:
            return [self.merge_aws_and_local_configuration(self.get_function_configuration(function_arn))
                    for function_arn in arn_list]
        except ClientError as cerr:
            print(f"Error getting function info by arn: {cerr}")

    def get_function_configuration(self, arn: str = None) -> Dict:
        function = arn if arn else self.function.get('name')
        return self.client.get_function_configuration(function)

    def get_fdl_config(self, arn: str = None) -> Dict:
        function = arn if arn else self.function.get('name')
        function_info = self.client.get_function(function)
        dep_pack_url = function_info.get('Code').get('Location')
        dep_pack = get_file(dep_pack_url)
        # Extract function_config.yaml
        try:
            with ZipFile(io.BytesIO(dep_pack)) as thezip:
                with thezip.open('function_config.yaml') as cfg_yaml:
                    return yaml.safe_load(cfg_yaml)
        except (KeyError, BadZipfile):
            return {}

    @excp.exception(logger)
    def find_function(self, function_name_or_arn=None):
        try:
            # If this call works the function exists
            name_arn = function_name_or_arn if function_name_or_arn else self.function.get('name', '')
            self.get_function_configuration(name_arn)
            return True
        except ClientError as ce:
            # Function not found
            if ce.response['Error']['Code'] == 'ResourceNotFoundException':
                return False
            else:
                raise

    def add_invocation_permission_from_api_gateway(self):
        api = self.resources_info.get('api_gateway')
        # Add Testing permission
        kwargs = {'FunctionName': self.function.get('name'),
                  'Principal': api.get('service_id'),
                  'SourceArn': api.get('source_arn_testing').format(api_region=api.get('region'),
                                                                    account_id=self.resources_info.get('iam').get('account_id'),
                                                                    api_id=api.get('id'))}
        self.client.add_invocation_permission(**kwargs)
        # Add Invocation permission
        kwargs['SourceArn'] = api.get('source_arn_invocation').format(api_region=api.get('region'),
                                                                       account_id=self.resources_info.get('iam').get('account_id'),
                                                                       api_id=api.get('id'),
                                                                       stage_name=api.get('stage_name'))
        self.client.add_invocation_permission(**kwargs)

    def get_api_gateway_id(self):
        env_vars = self._get_function_environment_variables()
        return env_vars['Variables'].get('API_GATEWAY_ID', '')

    def _get_api_gateway_url(self):
        api_id = self.get_api_gateway_id()
        if not api_id:
            raise excp.ApiEndpointNotFoundError(self.function.get('name'))
        return self.resources_info.get('api_gateway').get('endpoint').format(api_id=api_id,
                                                                             api_region=self.resources_info.get('api_gateway').get('region'),
                                                                             stage_name=self.resources_info.get('api_gateway').get('stage_name'))

    def call_http_endpoint(self):
        invoke_args = {'headers' : {'X-Amz-Invocation-Type':'Event'} if self.is_asynchronous() else {}}
        self._set_invoke_args(invoke_args)
        return call_http_endpoint(self._get_api_gateway_url(), **invoke_args)

    def _set_invoke_args(self, invoke_args):
        if self.resources_info.get('api_gateway').get('data_binary', False):
            invoke_args['data'] = self._get_b64encoded_binary_data()
            invoke_args['headers'].update({'Content-Type': 'application/octet-stream'})
        if self.resources_info.get('api_gateway').get('parameters', False):
            invoke_args['params'] = self._parse_http_parameters(self.resources_info.get('api_gateway').get('parameters'))
        if self.resources_info.get('api_gateway').get('json_data', False):
            invoke_args['data'] = self._parse_http_parameters(self.resources_info.get('api_gateway').get('json_data'))
            invoke_args['headers'].update({'Content-Type': 'application/json'})

    def _parse_http_parameters(self, parameters):
        return parameters if type(parameters) is dict else json.loads(parameters)

    @excp.exception(logger)
    def _get_b64encoded_binary_data(self):
        data_path = self.resources_info.get('api_gateway').get('data_binary')
        AWSValidator.validate_http_payload_size(data_path, self.is_asynchronous())
        with open(data_path, 'rb') as data_file:
            return base64.b64encode(data_file.read())



# import the necessary libs
import numpy as np
import argparse
import time
import cv2
import os

# construct the argument parser and parse the arguments
ap = argparse.ArgumentParser()
ap.add_argument("-i", "--image", required=True, help="path to input image")
ap.add_argument("-o", "--output",help="path to output image")
ap.add_argument("-y", "--yolo", default="/opt/mask-detector/cfg", help="base path to YOLO cfg directory")
ap.add_argument("-c", "--confidence", type=float, default=0.2, help="minimum probability to filter weak detections")
ap.add_argument("-t", "--threshold", type=float, default=0.1, help="threshold when applying non-max suppression")
args = vars(ap.parse_args())

# load the class labels our YOLO model was trained on
labelsPath = os.path.sep.join([args["yolo"], "obj.names"])
LABELS = open(labelsPath).read().strip().split("\n")

# initialize a list of colors to represent each possible class label (red and green)
COLORS = [[0,255,0], [0,0,255]]

# derive the paths to the YOLO weights and model configuration
weightsPath = os.path.sep.join([args["yolo"], "yolov3-tiny_obj_train_tiny8.weights"])
configPath = os.path.sep.join([args["yolo"], "yolov3-tiny_obj_train.cfg"])

# load our YOLO object detector 
print("[INFO] loading YOLO from disk...")
net = cv2.dnn.readNetFromDarknet(configPath, weightsPath)

# load our input image and get it height and width
image = cv2.imread(args["image"])
(H, W) = image.shape[:2]

# determine only the *output* layer names that we need from YOLO
ln = net.getLayerNames()

ln = [ln[i[0] - 1] for i in net.getUnconnectedOutLayers()]

# construct a blob from the input image and then perform a forward
# pass of the YOLO object detector, giving us our bounding boxes and
# associated probabilities
blob = cv2.dnn.blobFromImage(image, 1 / 255.0, (416, 416),swapRB=True, crop=False)
net.setInput(blob)
start = time.time()
layerOutputs = net.forward(ln) #list of 3 arrays, for each output layer.
end = time.time()

# show timing information on YOLO
print("[INFO] YOLO took {:.6f} seconds".format(end - start))

# initialize our lists of detected bounding boxes, confidences, and
# class IDs, respectively
boxes = []
confidences = []
classIDs = []

# loop over each of the layer outputs
for output in layerOutputs:

	# loop over each of the detections
	for detection in output:

		# extract the class ID and confidence (i.e., probability) of
		# the current object detection
		scores = detection[5:] #last 2 values in vector
		classID = np.argmax(scores)
		confidence = scores[classID]
		# filter out weak predictions by ensuring the detected
		# probability is greater than the minimum probability
		if confidence > args["confidence"]:
			# scale the bounding box coordinates back relative to the
			# size of the image, keeping in mind that YOLO actually
			# returns the center (x, y)-coordinates of the bounding
			# box followed by the boxes' width and height
			box = detection[0:4] * np.array([W, H, W, H])
			(centerX, centerY, width, height) = box.astype("int")
			# use the center (x, y)-coordinates to derive the top and
			# and left corner of the bounding box
			x = int(centerX - (width / 2))
			y = int(centerY - (height / 2))
			# update our list of bounding box coordinates, confidences,
			# and class IDs
			boxes.append([x, y, int(width), int(height)])
			confidences.append(float(confidence))
			classIDs.append(classID)

# apply NMS to suppress weak, overlapping bounding
# boxes
idxs = cv2.dnn.NMSBoxes(boxes, confidences, args["confidence"],args["threshold"])
border_size=100
border_text_color=[255,255,255]
#Add top-border to image to display stats
image = cv2.copyMakeBorder(image, border_size,0,0,0, cv2.BORDER_CONSTANT)
#calculate count values
filtered_classids=np.take(classIDs,idxs)
mask_count=(filtered_classids==0).sum()
nomask_count=(filtered_classids==1).sum()
#display count
text = "NoMaskCount: {}  MaskCount: {}".format(nomask_count, mask_count)
cv2.putText(image,text, (0, int(border_size-50)), cv2.FONT_HERSHEY_SIMPLEX,0.8,border_text_color, 2)
#display status
text = "Status:"
cv2.putText(image,text, (W-300, int(border_size-50)), cv2.FONT_HERSHEY_SIMPLEX,0.8,border_text_color, 2)
ratio=nomask_count/(mask_count+nomask_count)

if ratio>=0.1 and nomask_count>=3:
	text = "Danger !"
	cv2.putText(image,text, (W-200, int(border_size-50)), cv2.FONT_HERSHEY_SIMPLEX,0.8,[26,13,247], 2)
	
elif ratio!=0 and np.isnan(ratio)!=True:
	text = "Warning !"
	cv2.putText(image,text, (W-200, int(border_size-50)), cv2.FONT_HERSHEY_SIMPLEX,0.8,[0,255,255], 2)

else:
	text = "Safe "
	cv2.putText(image,text, (W-200, int(border_size-50)), cv2.FONT_HERSHEY_SIMPLEX,0.8,[0,255,0], 2)

# ensure at least one detection exists
if len(idxs) > 0:

	# loop over the indexes we are keeping
	for i in idxs.flatten():
		
		# extract the bounding box coordinates
		(x, y) = (boxes[i][0], boxes[i][1]+border_size)
		(w, h) = (boxes[i][2], boxes[i][3])
		# draw a bounding box rectangle and label on the image
		color = [int(c) for c in COLORS[classIDs[i]]]
		cv2.rectangle(image, (x, y), (x + w, y + h), color, 1)
		text = "{}: {:.4f}".format(LABELS[classIDs[i]], confidences[i])
		cv2.putText(image, text, (x, y-5), cv2.FONT_HERSHEY_SIMPLEX,0.5, color, 1)

if args["output"]:
	#save the image
	cv2.imwrite(args["output"],image)
else: 
	# show the output image
	cv2.imshow("Image",image)
	cv2.waitKey(0)


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
import yaml
from scar.providers.aws import GenericClient
import scar.logger as logger
from scar.providers.aws.launchtemplates import LaunchTemplates
from scar.providers.aws.functioncode import create_function_config
from scar.utils import FileUtils, StrUtils


def _get_job_definitions(jobs_info: Dict) -> List:
    return [f"{job_def.get('jobDefinitionName', '')}:{job_def.get('revision', '')}"
            for job_def in jobs_info.get('jobDefinitions', {})]


class Batch(GenericClient):

    def __init__(self, resources_info):
        super().__init__(resources_info.get('batch'))
        self.resources_info = resources_info
        self.batch = resources_info.get('batch')
        self.function_name = self.resources_info.get('lambda').get('name')

    def _set_required_environment_variables(self) -> None:
        self._set_batch_environment_variable('AWS_LAMBDA_FUNCTION_NAME', self.function_name)
        self._set_batch_environment_variable('SCRIPT', self._get_user_script())
        self._set_batch_environment_variable('FUNCTION_CONFIG', self._get_config_file())
        if self.resources_info.get('lambda').get('container').get('environment').get('Variables', False):
            for key, value in self.resources_info.get('lambda').get('container').get('environment').get('Variables').items():
                self._set_batch_environment_variable(key, value)

    def _set_batch_environment_variable(self, key: str, value: str) -> None:
        self.resources_info['batch']['environment']['Variables'].update({key: value})

    def _get_user_script(self) -> str:
        script = ''
        if self.resources_info.get('lambda').get('init_script', False):
            file_content = FileUtils.read_file(self.resources_info.get('lambda').get('init_script'))
            script = StrUtils.utf8_to_base64_string(file_content)
        return script

    def _get_config_file(self) -> str:
        cfg_file = ''
        config = create_function_config(self.resources_info)
        yaml_str = yaml.safe_dump(config)
        cfg_file = StrUtils.utf8_to_base64_string(yaml_str)
        return cfg_file

    def _delete_job_definitions(self) -> None:
        # Get main job definition
        kwargs = {"jobDefinitionName": self.function_name}
        job_info = self.client.describe_job_definitions(**kwargs)
        for job_def in _get_job_definitions(job_info):
            kwars = {"jobDefinition": job_def}
            self.client.deregister_job_definition(**kwars)
        logger.info("Job definitions successfully deleted.")

    def _get_job_queue_info(self):
        job_queue_info_args = {'jobQueues': [self.function_name]}
        return self.client.describe_job_queues(**job_queue_info_args)

    def _delete_job_queue(self):
        response = self._get_job_queue_info()
        while response["jobQueues"]:
            state = response["jobQueues"][0]["state"]
            status = response["jobQueues"][0]["status"]
            if status == "VALID":
                self._delete_valid_job_queue(state)
            response = self._get_job_queue_info()

    def _delete_valid_job_queue(self, state):
        if state == "ENABLED":
            updating_args = {'jobQueue': self.function_name,
                             'state': 'DISABLED'}
            self.client.update_job_queue(**updating_args)
        elif state == "DISABLED":
            deleting_args = {'jobQueue': self.function_name}
            self.client.delete_job_queue(**deleting_args)
            logger.info("Job queue successfully deleted.")

    def _get_describe_compute_env_args(self):
        return {'computeEnvironments': [self.function_name]}

    def _get_compute_env_info(self):
        creation_args = self._get_describe_compute_env_args()
        return self.client.describe_compute_environments(**creation_args)

    def _delete_compute_env(self):
        response = self._get_compute_env_info()
        while response["computeEnvironments"]:
            state = response["computeEnvironments"][0]["state"]
            status = response["computeEnvironments"][0]["status"]
            if status == "VALID":
                self._delete_valid_compute_environment(state)
            response = self._get_compute_env_info()

    def _delete_valid_compute_environment(self, state):
        if state == "ENABLED":
            update_args = {'computeEnvironment': self.function_name,
                           'state': 'DISABLED'}
            self.client.update_compute_environment(**update_args)
        elif state == "DISABLED":
            delete_args = {'computeEnvironment': self.function_name}
            self.client.delete_compute_environment(**delete_args)
            logger.info("Compute environment successfully deleted.")

    def _get_compute_env_args(self):
        account_id = self.resources_info.get('iam').get('account_id')
        return {
            'computeEnvironmentName': self.function_name,
            'serviceRole': self.batch.get('service_role').format(account_id=account_id),
            'type': self.batch.get('type'),
            'state':  self.batch.get('state'),
            'computeResources': {
                'type': self.batch.get('compute_resources').get('type'),
                'minvCpus': self.batch.get('compute_resources').get('min_v_cpus'),
                'maxvCpus': self.batch.get('compute_resources').get('max_v_cpus'),
                'desiredvCpus': self.batch.get('compute_resources').get('desired_v_cpus'),
                'instanceTypes': self.batch.get('compute_resources').get('instance_types'),
                'subnets': self.batch.get('compute_resources').get('subnets'),
                'securityGroupIds': self.batch.get('compute_resources').get('security_group_ids'),
                'instanceRole': self.batch.get('compute_resources').get('instance_role').format(account_id=account_id),
                'launchTemplate': {
                    'launchTemplateName': self.batch.get('compute_resources').get('launch_template_name'),
                    'version': str(LaunchTemplates(self.resources_info).get_launch_template_version())
                }
            }
        }

    def _get_creations_job_queue_args(self):
        return {
            'computeEnvironmentOrder': [{'computeEnvironment': self.function_name,
                                         'order': 1}, ],
            'jobQueueName':  self.function_name,
            'priority': 1,
            'state': self.batch.get('state'),
        }

    def _get_job_definition_args(self):
        job_def_args = {
            'jobDefinitionName': self.function_name
        }
        if self.batch.get('multi_node_parallel').get('enabled'):
            job_def_args['nodeProperties'] = self._get_node_properties_multi_node_args()
            job_def_args['type'] = 'multinode'
        else:
            job_def_args['containerProperties'] = self._get_container_properties_single_node_args()
            job_def_args['type'] = 'container'
        return job_def_args

    def _get_container_properties_single_node_args(self):
        job_def_args = {
            'image': self.resources_info.get('lambda').get('container').get('image'),
            'memory': int(self.batch.get('memory')),
            'vcpus': int(self.batch.get('vcpus')),
            'command': [
                '/bin/sh',
                '-c',
                'echo $EVENT | /opt/faas-supervisor/bin/supervisor'
            ],
            'volumes': [
                {
                    'host': {
                        'sourcePath': '/opt/faas-supervisor/bin'
                    },
                    'name': 'supervisor-bin'
                }
            ],
            'environment': [{'name': key, 'value': value} for key, value in self.resources_info['batch']['environment']['Variables'].items()],
            'mountPoints': [
                {
                    'containerPath': '/opt/faas-supervisor/bin',
                    'sourceVolume': 'supervisor-bin'
                }
            ]
        }
        if self.batch.get('enable_gpu'):
            job_def_args['containerProperties']['resourceRequirements'] = [
                {
                    'value': '1',
                    'type': 'GPU'
                }
            ]
        return job_def_args

    def _get_node_properties_multi_node_args(self):
        targetNodes = self.batch.get('multi_node_parallel').get('number_nodes') - 1
        job_def_args = {
            "numNodes": int(self.batch.get('multi_node_parallel').get('number_nodes')),
            "mainNode": int(self.batch.get('multi_node_parallel').get('main_node_index')),
            "nodeRangeProperties": [
                {
                "targetNodes": "0:", #+ str(targetNodes),
                "container": self._get_container_properties_single_node_args()
                }
            ]#[self._get_node_node_range_property_multi_node_args(target_nodes) for target_nodes in self.batch.get('multi_node_parallel').get('target_nodes')]
        }
        return job_def_args

    def _get_state_and_status_of_compute_env(self):
        creation_args = self._get_describe_compute_env_args()
        response = self.client.describe_compute_environments(**creation_args)
        return (response["computeEnvironments"][0]["state"],
                response["computeEnvironments"][0]["status"])

    def create_batch_environment(self):
        self._set_required_environment_variables()
        creation_args = self._get_compute_env_args()
        self.client.create_compute_environment(**creation_args)
        while True:
            state, status = self._get_state_and_status_of_compute_env()
            if state == "ENABLED" and status == "VALID":
                creation_args = self._get_creations_job_queue_args()
                logger.info("Compute environment successfully created.")
                self.client.create_job_queue(**creation_args)
                logger.info('Job queue successfully created.')
                creation_args = self._get_job_definition_args()
                logger.info(f"Registering '{self.function_name}' job definition.")
                return self.client.register_job_definition(**creation_args)

    def delete_compute_environment(self):
        self._delete_job_definitions()
        self._delete_job_queue()
        self._delete_compute_env()

    def exist_compute_environments(self):
        creation_args = self._get_describe_compute_env_args()
        response = self.client.describe_compute_environments(**creation_args)
        return len(response["computeEnvironments"]) > 0

    def get_jobs_with_request_id(self) -> Dict:
        describe_args = {'jobs': [self.resources_info.get('cloudwatch').get('request_id')]}
        return self.client.describe_jobs(**describe_args)

#     def exist_job(self, job_id: str) -> bool:
#         response = self.describe_jobs(job_id)
#         return len(response["jobs"]) != 0



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes in charge
of managing the SCAR configuration file."""

import json
import os
from scar.exceptions import exception, ScarConfigFileError
import scar.logger as logger
from scar.utils import FileUtils, SysUtils, StrUtils

_DEFAULT_CFG = {
    "scar": {
        "config_version": "1.1.0"
    },
    "oscar": {
        "my_oscar": {
            # Cluster credentials
            "endpoint": "",
            "auth_user": "",
            "auth_password": "",
            "ssl_verify": True,
            # Default service parameters
            # Memory limit for the service following the kubernetes format
            # https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container/#meaning-of-memory
            "memory": "256Mi",
            # CPU limit for the service following the kubernetes format
	        # https://kubernetes.io/docs/concepts/configuration/manage-compute-resources-container/#meaning-of-cpu
            "cpu": "0.2",
            "log_level": "INFO",
        }
    },
    "aws": {
        "iam": {"boto_profile": "default",
                "role": ""},
        "lambda": {
            "boto_profile": "default",
            "region": "us-east-1",
            "execution_mode": "lambda",
            "timeout": 300,
            "memory": 512,
            "description": "Automatically generated lambda function",
            "runtime": "python3.7",
            "layers": [],
            "invocation_type": "RequestResponse",
            "asynchronous": False,
            "log_type": "Tail",
            "log_level": "INFO",
            "environment": {
                "Variables": {
                    "UDOCKER_BIN" : "/opt/udocker/bin/",
                    "UDOCKER_LIB" : "/opt/udocker/lib/",
                    "UDOCKER_DIR" : "/tmp/shared/udocker",
                    "UDOCKER_EXEC": "/opt/udocker/udocker.py"}},
            "deployment": {
                "max_payload_size": 52428800,
                "max_s3_payload_size": 262144000           
            },
            "container": {
                "environment" : {
                    "Variables" : {}},
                "timeout_threshold": 10
            },
            # Must be a Github tag or "latest"
            "supervisor": {
                "version": "latest",
                'layer_name' : "faas-supervisor",
                'license_info' : 'Apache 2.0'                
            }
        },
        "s3": {
            "boto_profile": "default",
            "region": "us-east-1",            
            "event" : {
                "Records": [{
                    "eventSource": "aws:s3",
                    "s3" : {
                        "bucket" : {
                            "name": "{bucket_name}",
                            "arn": "arn:aws:s3:::{bucket_name}"
                        },
                        "object" : {
                            "key": "{file_key}"
                        }
                    }
                }]
            }
        },
        "api_gateway": {
            "boto_profile": "default",
            "region": "us-east-1",            
            "endpoint": "https://{api_id}.execute-api.{api_region}.amazonaws.com/{stage_name}/launch",
            'request_parameters': {"integration.request.header.X-Amz-Invocation-Type":
                                   "method.request.header.X-Amz-Invocation-Type"},
            # ANY, DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT
            'http_method': "ANY",
            "method" : {
                # NONE, AWS_IAM, CUSTOM, COGNITO_USER_POOLS
                'authorizationType': "NONE",
                'requestParameters': {'method.request.header.X-Amz-Invocation-Type' : False},
            },
            "integration": {
                # 'HTTP'|'AWS'|'MOCK'|'HTTP_PROXY'|'AWS_PROXY'
                'type': "AWS_PROXY",
                'integrationHttpMethod' : "POST",
                'uri' : "arn:aws:apigateway:{api_region}:lambda:path/2015-03-31/functions/arn:aws:lambda:{lambda_region}:{account_id}:function:{function_name}/invocations",
                'requestParameters' : {"integration.request.header.X-Amz-Invocation-Type":
                                       "method.request.header.X-Amz-Invocation-Type"}            
            },
            'path_part': "{proxy+}",
            'stage_name': "scar",
            # Used to add invocation permissions to lambda
            'service_id': 'apigateway.amazonaws.com',
            'source_arn_testing': 'arn:aws:execute-api:{api_region}:{account_id}:{api_id}/*',
            'source_arn_invocation': 'arn:aws:execute-api:{api_region}:{account_id}:{api_id}/{stage_name}/ANY'
        },
        "cloudwatch": {
            "boto_profile": "default",
            "region": "us-east-1",
            "log_retention_policy_in_days": 30
        },
        "batch": {
            "multi_node_parallel": {
                "enabled": False,
                "number_nodes": 10,
                "main_node_index": 0
            },
            "boto_profile": "default",
            "region": "us-east-1",
            "vcpus": 1,
            "memory": 1024,
            "enable_gpu": False,
            "state": "ENABLED",
            "type": "MANAGED",
            "environment" : {
                "Variables" : {}},
            "compute_resources": {
                "security_group_ids": [],
                "type": "EC2",
                "desired_v_cpus": 0,
                "min_v_cpus": 0,
                "max_v_cpus": 2,
                "subnets": [],
                "instance_types": ["m3.medium"],
                "launch_template_name": "faas-supervisor",
                "instance_role": "arn:aws:iam::{account_id}:instance-profile/ecsInstanceRole"
            },
            "service_role": "arn:aws:iam::{account_id}:role/service-role/AWSBatchServiceRole"            
        }
    }
}


class ConfigFileParser():
    """Class to manage the SCAR configuration file creation, update and load."""

    _CONFIG_FOLDER_PATH_ENV_VAR = 'SCAR_CONFIG_FOLDER'
    _CONFIG_FOLDER_PATH = ".scar"
    _CONFIG_FILE_PATH = "scar.cfg"
    _CONFIG_FILE_NAME_BCK = "scar.cfg_old"
    _CONFIG_FILE_NAME_TMP_YAML = "scar_tmp.yaml"

    # Set default config folder path
    config_file_folder = FileUtils.join_paths(SysUtils.get_user_home_path(), _CONFIG_FOLDER_PATH)

    # Check if config folder env var is set, and use it for config paths
    if _CONFIG_FOLDER_PATH_ENV_VAR in os.environ:
        config_file_folder = os.getenv(_CONFIG_FOLDER_PATH_ENV_VAR)

    # Set config file paths
    config_file_path = FileUtils.join_paths(config_file_folder, _CONFIG_FILE_PATH)
    backup_file_path = FileUtils.join_paths(config_file_folder, _CONFIG_FILE_NAME_BCK)
    tmp_yaml_file_path = FileUtils.join_paths(config_file_folder, _CONFIG_FILE_NAME_TMP_YAML)

    @exception(logger)
    def __init__(self):
        # Check if the config file exists
        if FileUtils.is_file(self.config_file_path):
            with open(self.config_file_path) as cfg_file:
                self.cfg_data = json.load(cfg_file)
            if not self._is_config_file_updated():
                self._update_config_file()
        else:
            self._create_scar_config_folder_and_file()

    def _is_config_file_updated(self):
        if 'config_version' not in self.cfg_data['scar']:
            return False
        return StrUtils.compare_versions(self.cfg_data.get('scar', {}).get("config_version", ""),
                                         _DEFAULT_CFG['scar']["config_version"]) >= 0

    def get_properties(self):
        """Returns the configuration data of the configuration file."""
        return self.cfg_data

    def get_udocker_zip_url(self):
        """Returns the url where the udocker zip is stored."""
        return self.cfg_data['scar']['udocker_info']['zip_url']

    def _create_scar_config_folder_and_file(self):
        FileUtils.create_folder(self.config_file_folder)
        self._create_new_config_file()
        raise ScarConfigFileError(file_path=self.config_file_path)

    def _create_new_config_file(self):
        FileUtils.create_file_with_content(self.config_file_path,
                                           json.dumps(_DEFAULT_CFG, indent=2))

    def _update_config_file(self):
        logger.info(("SCAR configuration file deprecated.\n"
                     "Updating your SCAR configuration file."))
        FileUtils.copy_file(self.config_file_path, self.backup_file_path)
        logger.info(f"Old configuration file saved in '{self.backup_file_path}'.")
        self._create_new_config_file()
        logger.info((f"New configuration file saved in '{self.config_file_path}'.\n"
                     "Please fill your new configuration file with your account information."))
        SysUtils.finish_scar_execution()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict
from scar.utils import DataTypesUtils

FAAS_PROVIDERS = ["aws", "oscar"]

def merge_conf(conf: Dict, yaml: Dict) -> Dict:
    result = yaml.copy()
    # We have to merge the default config with all the defined functions
    for provider in FAAS_PROVIDERS:
        for index, function in enumerate(result.get('functions', {}).get(provider, {})):
            result['functions'][provider][index] = \
                DataTypesUtils.merge_dicts_with_copy(conf.get(provider,{}), function)
    result['scar'] = DataTypesUtils.merge_dicts_with_copy(result.get('scar', {}),
                                                          conf.get('scar', {}))
    return result

def merge_cmd_yaml(cmd: Dict, yaml: Dict) -> Dict:
    result = yaml.copy()
    # We merge the cli commands with all the defined functions
    # CLI only allows define AWS parameters
    for cli_cmd in cmd.get('functions', {}).get("aws", {}):
        for index, function in enumerate(result.get('functions', {}).get("aws", {})):
            result['functions']['aws'][index] = \
                DataTypesUtils.merge_dicts_with_copy(function, cli_cmd)
    result['scar'] = DataTypesUtils.merge_dicts_with_copy(result.get('scar', {}),
                                                          cmd.get('scar', {}))
    result['storage_providers'] = DataTypesUtils.merge_dicts_with_copy(result.get('storage_providers', {}),
                                                                       cmd.get('storage_providers', {}))
    return result



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
import abc

class CallType(Enum):
    INIT = "init"
    INVOKE = "invoke"
    RUN = "run"
    LS = "ls"
    RM = "rm"
    LOG = "log"
    PUT = "put"
    GET = "get"  

class Commands(metaclass=abc.ABCMeta):
    ''' All the different cloud provider controllers must inherit 
    from this class to ensure that the commands are defined consistently'''

    @abc.abstractmethod
    def init(self):
        pass

    @abc.abstractmethod    
    def invoke(self):
        pass

    @abc.abstractmethod    
    def run(self):
        pass

    @abc.abstractmethod    
    def ls(self):
        pass

    @abc.abstractmethod
    def rm(self):
        pass

    @abc.abstractmethod
    def log(self):
        pass

    @abc.abstractmethod
    def put(self):
        pass

    @abc.abstractmethod
    def get(self):
        pass    



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zipfile import ZipFile
from scar.utils import FileUtils, SysUtils, StrUtils


def _extract_udocker_zip(supervisor_zip_path) -> None:
    file_path = ""
    with ZipFile(supervisor_zip_path) as thezip:
        for file in thezip.namelist():
            if file.endswith("udocker.zip"):
                file_path = FileUtils.join_paths(FileUtils.get_tmp_dir(), file)
                thezip.extract(file, FileUtils.get_tmp_dir())
                break
    return file_path


class Udocker():

    _CONTAINER_NAME = "udocker_container"

    def __init__(self, resources_info: str, tmp_payload_folder_path: str, supervisor_zip_path: str):
        self.resources_info = resources_info
        self._tmp_payload_folder_path = tmp_payload_folder_path
        self._udocker_dir = FileUtils.join_paths(self._tmp_payload_folder_path, "udocker")
        self._udocker_dir_orig = ""
        self._udocker_code = FileUtils.join_paths(self._udocker_dir, "udocker.py")
        self._udocker_exec = ['python3', self._udocker_code]
        self._install_udocker(supervisor_zip_path)

    def _install_udocker(self, supervisor_zip_path: str) -> None:
        udocker_zip_path = _extract_udocker_zip(supervisor_zip_path)
        with ZipFile(udocker_zip_path) as thezip:
            thezip.extractall(self._tmp_payload_folder_path)

    def _save_tmp_udocker_env(self):
        # Avoid override global variables
        if SysUtils.is_variable_in_environment("UDOCKER_DIR"):
            self._udocker_dir_orig = SysUtils.get_environment_variable("UDOCKER_DIR")
        # Set temporal global vars
        SysUtils.set_environment_variable("UDOCKER_DIR", self._udocker_dir)

    def _restore_udocker_env(self):
        if self._udocker_dir_orig:
            SysUtils.set_environment_variable("UDOCKER_DIR", self._udocker_dir_orig)
        else:
            SysUtils.delete_environment_variable("UDOCKER_DIR")

    def _set_udocker_local_registry(self):
        self.resources_info['lambda']['environment']['Variables']['UDOCKER_REPOS'] = '/var/task/udocker/repos/'
        self.resources_info['lambda']['environment']['Variables']['UDOCKER_LAYERS'] = '/var/task/udocker/layers/'


    def prepare_udocker_image(self):
        self._save_tmp_udocker_env()
        cmd_out = SysUtils.execute_command_with_msg(self._udocker_exec + ["load", "-i",
                                                                          self.resources_info.get('lambda').get('container').get('image_file')],
                                                    cli_msg="Loading image file")
        # Get the image name from the command output
        self.resources_info['lambda']['container']['image'] = cmd_out.split('\n')[1]
        self._set_udocker_local_registry()
        self._restore_udocker_env()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class implementing the low-level functions to 
communicate with an OSCAR cluster."""

from typing import Dict, List
from tabulate import tabulate
from scar.providers.aws.response import OutputType
import scar.logger as logger
from scar.utils import StrUtils


def parse_ls_response(oscar_resources: List, endpoint: str, cluster_id: str, output_type: int) -> None:
    oscar_output = 'Services'
    result = []
    text_message = f'\nOSCAR SERVICES - CLUSTER "{cluster_id}" ({endpoint}):\n'
    for resources_info in oscar_resources:
        result.append(_parse_service_info(resources_info))
    text_message += _get_table(result)
    _print_generic_response({oscar_output: oscar_resources}, output_type, text_message, {oscar_output: result})


def parse_service_creation(resources_info: Dict, output_type: int) -> None:
    result = _parse_service_info(resources_info)
    text_message = f'Service \'{resources_info["name"]}\' successfully created on cluster \'{resources_info.get("cluster_id")}\'.'
    _print_generic_response(resources_info, output_type, text_message, result)


def parse_service_deletion(resources_info: Dict, output_type: int) -> None:
    result = _parse_service_info(resources_info)
    text_message = f'Service \'{resources_info.get("name")}\' successfully deleted from cluster \'{resources_info.get("cluster_id")}\'.'
    _print_generic_response(resources_info, output_type, text_message, result)


def _parse_service_info(resources_info: Dict) -> Dict:
    return {'Name': resources_info.get('name', '-'),
            'Memory': resources_info.get('memory', '-'),
            'CPU': resources_info.get('cpu', '-'),
            'Image_id': resources_info.get('image', '-')}


def _get_table(services_info: List) -> str:
    headers = ['NAME', 'MEMORY', 'CPU', 'IMAGE_ID']
    table = []
    for function in services_info:
        table.append([function['Name'],
                      function['Memory'],
                      function['CPU'],
                      function['Image_id']])
    return tabulate(table, headers)


def _print_generic_response(resources_info: Dict, output_type: int, text_message: str, json_output: Dict) -> None:
    # Support 'PLAIN_TEXT', 'JSON' and 'VERBOSE' output types
    if output_type == OutputType.PLAIN_TEXT.value:
        output = text_message
        logger.info(output)
    else:
        if output_type == OutputType.JSON.value:
            output = json_output
        elif output_type == OutputType.VERBOSE.value:
            output = resources_info
        logger.info_json(output)




#! /usr/bin/python

# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
sys.path.append('.')

from scar.parser.cfgfile import ConfigFileParser
from scar.parser.cli import CommandParser
from scar.providers.aws.controller import AWS
from scar.providers.oscar.controller import OSCAR
from scar.utils import FileUtils
import scar.parser.fdl as fdl
import scar.exceptions as excp
import scar.logger as logger


@excp.exception(logger)
def parse_arguments():
    """
    Merge the scar.conf parameters, the cmd parameters and the yaml
    file parameters in a single dictionary.

    The precedence of parameters is CMD >> YAML >> SCAR.CONF
    That is, the CMD parameter will override any other configuration,
    and the YAML parameters will override the SCAR.CONF settings
    """
    config_args = ConfigFileParser().get_properties()
    func_call, cmd_args = CommandParser().parse_arguments()
    if 'conf_file' in cmd_args['scar'] and cmd_args['scar']['conf_file']:
        yaml_args = FileUtils.load_yaml(cmd_args['scar']['conf_file'])
        # YAML >> SCAR.CONF
        merged_args = fdl.merge_conf(config_args, yaml_args)
        merged_args = fdl.merge_cmd_yaml(cmd_args, merged_args)
    else:
        # CMD >> SCAR.CONF
        merged_args = fdl.merge_conf(config_args, cmd_args)
    #self.cloud_provider.parse_arguments(merged_args)
    FileUtils.create_tmp_config_file(merged_args, ConfigFileParser())
    return func_call

def main():
    logger.init_execution_trace()
    try:
        func_call = parse_arguments()
        # Default provider
        # If more providers, analyze the arguments and build the required one
        AWS(func_call)
        # Build the OSCAR controller only with 'init', 'rm' and 'ls' commands
        if func_call in ['init', 'rm', 'ls']:
            OSCAR(func_call)
        logger.end_execution_trace()
    except Exception as excp:
        print(excp)
        logger.exception(excp)
        logger.end_execution_trace_with_errors()


if __name__ == "__main__":
    main()



import tensorflow as tf
import numpy as np
import time

class DetectorAPI:
    def __init__(self, path_to_ckpt):
        self.path_to_ckpt = path_to_ckpt

        self.detection_graph = tf.Graph()
        with self.detection_graph.as_default():
            od_graph_def = tf.compat.v1.GraphDef()
            with tf.io.gfile.GFile(self.path_to_ckpt, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')

        self.default_graph = self.detection_graph.as_default()
        self.sess = tf.compat.v1.Session(graph=self.detection_graph)

        # Definite input and output Tensors for detection_graph
        self.image_tensor = self.detection_graph.get_tensor_by_name(
            'image_tensor:0')
        # Each box represents a part of the image where a particular object was detected.
        self.detection_boxes = self.detection_graph.get_tensor_by_name(
            'detection_boxes:0')
        # Each score represent how level of confidence for each of the objects.
        # Score is shown on the result image, together with the class label.
        self.detection_scores = self.detection_graph.get_tensor_by_name(
            'detection_scores:0')
        self.detection_classes = self.detection_graph.get_tensor_by_name(
            'detection_classes:0')
        self.num_detections = self.detection_graph.get_tensor_by_name(
            'num_detections:0')

    def processFrame(self, image):
        # Expand dimensions since the trained_model expects images to have shape: [1, None, None, 3]
        image_np_expanded = np.expand_dims(image, axis=0)
        # Actual detection.
        start_time = time.time()
        (boxes, scores, classes,
         num) = self.sess.run([
             self.detection_boxes, self.detection_scores,
             self.detection_classes, self.num_detections
         ],
            feed_dict={self.image_tensor: image_np_expanded})
        end_time = time.time()

        print("Elapsed Time:", end_time - start_time)

        im_height, im_width, _ = image.shape
        boxes_list = [None for i in range(boxes.shape[1])]
        for i in range(boxes.shape[1]):
            boxes_list[i] = (int(boxes[0, i, 1] * im_width),
                             int(boxes[0, i, 0] * im_height),
                             int(boxes[0, i, 3] * im_width),
                             int(boxes[0, i, 2] * im_height))

        return boxes_list, scores[0].tolist(), [
            int(x) for x in classes[0].tolist()
        ], int(num[0])

    def close(self):
        self.sess.close()
        self.default_graph.close()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the generic classes and methods used in the boto client modules."""

from typing import Dict
import boto3
import botocore
from scar.utils import lazy_property


class BotoClient():
    """Generic class in charge of creating a boto client based on the
    'boto_client_name' property of the class that inherits it."""

    _READ_TIMEOUT = 360
    _BOTO_CLIENT_NAME = ''

    def __init__(self, client_args: Dict):
        self.client_args = client_args.get('client', {})
        self.session_args = client_args.get('session', {})

    @lazy_property
    def client(self):
        """Returns a boto client based on the 'boto_client_name' property,
        the region specified on the client args and
        the profile specified on the session args."""
        # 'default' profile if nothing set
        session = boto3.Session(**self.session_args)
        self.client_args['config'] = botocore.config.Config(read_timeout=self._READ_TIMEOUT)
        return session.client(self._BOTO_CLIENT_NAME, **self.client_args)

    def get_access_key(self) -> str:
        """Returns the access key belonging to the boto_profile used."""
        session = boto3.Session(**self.session_args)
        credentials = session.get_credentials()
        return credentials.access_key



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with methods and classes in charge
of parsing the SCAR CLI commands."""

import argparse
import sys
from typing import Dict, List
from scar.parser.cli.subparsers import Subparsers
from scar.parser.cli.parents import *
from scar.utils import DataTypesUtils, StrUtils, FileUtils
from scar.cmdtemplate import CallType
import scar.exceptions as excp
import scar.logger as logger
import scar.version as version


def _parse_aws_args(cmd_args: Dict) -> Dict:
    aws_args = {}
    other_args = [('profile', 'boto_profile'), 'region', 'execution_mode']
    _set_args(aws_args, 'iam', _parse_iam_args(cmd_args))
    _set_args(aws_args, 'lambda', _parse_lambda_args(cmd_args))
    _set_args(aws_args, 'batch', _parse_batch_args(cmd_args))
    _set_args(aws_args, 'cloudwatch', _parse_cloudwatchlogs_args(cmd_args))
    _set_args(aws_args, 'api_gateway', _parse_api_gateway_args(cmd_args))
    aws_args.update(DataTypesUtils.parse_arg_list(other_args, cmd_args))
    storage = _parse_s3_args(aws_args, cmd_args)
    result = {'functions': {'oscar': [{}], 'aws': [aws_args]}}
    if storage:
        result.update(storage)
    return result


def _set_args(args: Dict, key: str, val: str) -> None:
    if key and val:
        args[key] = val


def _parse_scar_args(cmd_args: Dict) -> Dict:
    scar_args = ['conf_file', 'json', 'verbose', 'path', 'execution_mode',
                 'output_file', 'supervisor_version', 'all']
    return {'scar' : DataTypesUtils.parse_arg_list(scar_args, cmd_args)}


def _parse_iam_args(cmd_args: Dict) -> Dict:
    iam_args = [('iam_role', 'role')]
    return DataTypesUtils.parse_arg_list(iam_args, cmd_args)


def _parse_lambda_args(cmd_args: Dict) -> Dict:
    lambda_arg_list = ['name', 'asynchronous', 'init_script', 'run_script', 'c_args', 'memory',
                       'timeout', 'timeout_threshold', 'image', 'image_file', 'description',
                       'lambda_role', 'extra_payload', ('environment', 'environment_variables'),
                       'layers', 'lambda_environment', 'list_layers', 'log_level', 'preheat']
    lambda_args = DataTypesUtils.parse_arg_list(lambda_arg_list, cmd_args)
    # Standardize log level if defined
    if "log_level" in lambda_args:
        lambda_args['log_level'] = lambda_args['log_level'].upper()
    # Parse environment variables
    lambda_args.update(_get_lambda_environment_variables(lambda_args))
    return lambda_args


def _get_lambda_environment_variables(lambda_args: Dict) -> None:
    lambda_env_vars = {"environment": {"Variables": {}},
                       "container": {'environment' : {"Variables": {}}}}
    if "environment_variables" in lambda_args:
        # These variables define the udocker container environment variables
        for env_var in lambda_args["environment_variables"]:
            key_val = env_var.split("=")
            # Add an specific prefix to be able to find the container variables defined by the user
            lambda_env_vars['container']['environment']['Variables'][f'{key_val[0]}'] = key_val[1]
        del(lambda_args['environment_variables'])
    if "extra_payload" in lambda_args:
        lambda_env_vars['container']['extra_payload'] = f"/var/task"
    if "init_script" in lambda_args:
        lambda_env_vars['container']['init_script'] = f"/var/task/{FileUtils.get_file_name(lambda_args['init_script'])}"
    if "image" in lambda_args:
        lambda_env_vars['container']['image'] = lambda_args.get('image')
        del(lambda_args['image'])
    if "image_file" in lambda_args:
        lambda_env_vars['container']['image_file'] = lambda_args.get('image_file')
        del(lambda_args['image_file'])        

    if "lambda_environment" in lambda_args:
        # These variables define the lambda environment variables
        for env_var in lambda_args["lambda_environment"]:
            key_val = env_var.split("=")
            lambda_env_vars['environment']['Variables'][f'{key_val[0]}'] = key_val[1]
        del(lambda_args['lambda_environment'])
    return lambda_env_vars


def _parse_batch_args(cmd_args: Dict) -> Dict:
    batch_args = [('batch_vcpus', 'vcpus'), ('batch_memory', 'memory'), 'enable_gpu']
    return DataTypesUtils.parse_arg_list(batch_args, cmd_args)


def _parse_cloudwatchlogs_args(cmd_args: Dict) -> Dict:
    cw_log_args = ['log_stream_name', 'request_id']
    return DataTypesUtils.parse_arg_list(cw_log_args, cmd_args)


def _parse_s3_args(aws_args: Dict, cmd_args: Dict) -> Dict:
    s3_arg_list = ['deployment_bucket',
                   'input_bucket',
                   'output_bucket',
                   ('bucket', 'input_bucket')]

    s3_args = DataTypesUtils.parse_arg_list(s3_arg_list, cmd_args)
    storage = {}
    if s3_args:
        if 'deployment_bucket' in s3_args:
            aws_args['lambda']['deployment'] = {'bucket': s3_args['deployment_bucket']}
        if 'input_bucket' in s3_args:
            aws_args['lambda']['input'] = [{'storage_provider': 's3', 'path':  s3_args['input_bucket']}]
        if 'output_bucket' in s3_args:
            aws_args['lambda']['output'] = [{'storage_provider': 's3', 'path':  s3_args['output_bucket']}]
        storage['storage_providers'] = {'s3': {}}
    return storage


def _parse_api_gateway_args(cmd_args: Dict) -> Dict:
    api_gtw_args = [('api_gateway_name', 'name'), 'parameters', 'data_binary', 'json_data']
    return DataTypesUtils.parse_arg_list(api_gtw_args, cmd_args)


def _create_main_parser():
    parser = argparse.ArgumentParser(prog="scar",
                                     description=("Deploy containers "
                                                  "in serverless architectures"),
                                     epilog=("Run 'scar COMMAND --help' "
                                             "for more information on a command."))
    parser.add_argument('--version',
                        help='Show SCAR version.',
                        dest="version",
                        action="store_true",
                        default=False)
    return parser


def _create_parent_parsers() -> Dict:
    parsers = {}
    parsers['function_definition_parser'] = create_function_definition_parser()
    parsers['exec_parser'] = create_exec_parser()
    parsers['output_parser'] = create_output_parser()
    parsers['profile_parser'] = create_profile_parser()
    parsers['storage_parser'] = create_storage_parser()
    return parsers


class CommandParser():
    """Class to manage the SCAR CLI commands."""

    def __init__(self):
        self.parser = _create_main_parser()
        self.parent_parsers = _create_parent_parsers()
        self._add_subparsers()

    def _add_subparsers(self) -> None:
        subparsers = Subparsers(self.parser.add_subparsers(title='Commands'), self.parent_parsers)
        # We need to define a subparser for each command defined in the CallType class
        for cmd in CallType:
            subparsers.add_subparser(cmd.value)

    @excp.exception(logger)
    def parse_arguments(self) -> Dict:
        """Command parsing and selection"""
        try:
            cmd_args = self.parser.parse_args()
            if cmd_args.version:
                print(f"SCAR {version.__version__}")
                sys.exit(0)
            cmd_args = vars(cmd_args)
            if 'func' not in cmd_args:
                raise excp.MissingCommandError()
            scar_args = _parse_scar_args(cmd_args)
            aws_args = _parse_aws_args(cmd_args)
            return cmd_args['func'], DataTypesUtils.merge_dicts_with_copy(scar_args, aws_args)
        except AttributeError as aerr:
            logger.error("Incorrect arguments: use scar -h to see the options available",
                         f"Error parsing arguments: {aerr}")
        except:
            print("Unexpected error:", sys.exc_info()[0])
            raise



# author: Asmaa Mirkhan ~ 2019

import os
import argparse
import cv2 as cv


def blurBoxes(image, boxes):
    """
    Argument:
    image -- the image that will be edited as a matrix
    boxes -- list of boxes that will be blurred, each box must be int the format (x_top_left, y_top_left, width, height)
    
    Returns:
    image -- the blurred image as a matrix
    """
    
    for box in boxes:
        # unpack each box
        x,y,w,h = [d for d in box]
        
        # crop the image due to the current box
        sub = image[y:y+h, x:x+w]
        
        # apply GaussianBlur on cropped area
        blur = cv.GaussianBlur(sub, (23,23), 30)
        
        # paste blurred image on the original image
        image[y:y+h, x:x+w] = blur
        
    return image


def main(args):
    # open the image
    image = cv.imread(args.input_image)
    
    # create a copy and do temp operations without affecting the original image
    temp_image = image.copy()
    
    # an array to store selected regions coordinates
    ROIs = []
    
    # keep getting ROIs until pressing 'q'
    while True:
        # get ROI cv.selectROI(window_name, image_matrix, selecting_start_point) 
        box = cv.selectROI('blur', temp_image, fromCenter=False)
        
        # add selected box to box list
        ROIs.append(box)
        
        # draw a rectangle on selected ROI
        cv.rectangle(temp_image, (box[0],box[1]), (box[0]+box[2], box[1]+box[3]), (0,255,0), 3)
        print('ROI is saved, press q to stop capturing, press any other key to select other ROI')
        
        # if 'q' is pressed then break
        key = cv.waitKey(0)
        if key & 0xFF == ord('q'):
            break
    
    # apply blurring
    image = blurBoxes(image, ROIs)
    
    # if image will be saved then save it
    if args.output_image:
        cv.imwrite(args.output_image,image)
    cv.imshow('blurred',image)
    cv.waitKey(0)
    
    

if __name__ == "__main__":
    # creating argument parser
    parser = argparse.ArgumentParser(description='Image blurring parameters')
    
    # adding arguments
    parser.add_argument('-i', '--input_image',
                        help='Path to your image', type=str, required=True)
    parser.add_argument('-o', '--output_image',
                        help='Output file path', type=str)
    args = parser.parse_args()
    
    # if input image path is invalid then stop
    assert os.path.isfile(args.input_image), 'Invalid input file'
    
    # if output directory is invalid then stop
    if args.output_image:
        assert os.path.isdir(os.path.dirname(
            args.output_image)), 'No such directory'

    main(args)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse

FUNCTION_DEFINITION = "function_definition_parser"
OUTPUT = "output_parser"
PROFILE = "profile_parser"
EXEC = "exec_parser"
STORAGE = "storage_parser"

INIT_PARENTS = [PROFILE, FUNCTION_DEFINITION, OUTPUT]
INVOKE_PARENTS = [PROFILE, EXEC]
RUN_PARENTS = [PROFILE, EXEC, OUTPUT]
RM_LS_PARENTS = [PROFILE, OUTPUT]
LOG_PARENTS = [PROFILE]
PUT_GET_PARENTS = [PROFILE, STORAGE]


class Subparsers():

    def __init__(self, subparser, parents):
        self.subparser = subparser
        self.parent_parsers = parents

    def _get_parents(self, parent_sublist):
        return [self.parent_parsers.get(parent, "") for parent in parent_sublist]

    def add_subparser(self, name):
        getattr(self, f'_add_{name}_parser')()

    def _add_init_parser(self):
        init = self.subparser.add_parser('init',
                                         parents=self._get_parents(INIT_PARENTS),
                                         help="Create lambda function")
        # Set default function
        init.set_defaults(func="init")
        # Lambda conf
        group = init.add_mutually_exclusive_group(required=True)
        group.add_argument("-i", "--image",
                           help="Container image id (i.e. centos:7)")
        group.add_argument("-if", "--image-file",
                           help=("Container image file created with "
                                 "'docker save' (i.e. centos.tar.gz)"))
        group.add_argument("-f", "--conf-file",
                           help="Yaml file with the function configuration")
        init.add_argument("-n", "--name", help="Lambda function name")
        init.add_argument("-s", "--init-script", help=("Path to the input file "
                                                       "passed to the function"))
        init.add_argument("-ph", "--preheat",
                          help=("Invokes the function once and downloads the container"),
                          action="store_true")
        init.add_argument("-ep", "--extra-payload",
                          help=("Folder containing files that are going to be "
                                "added to the lambda function"))
        init.add_argument("-db", "--deployment-bucket",
                          help="Bucket where the deployment package is going to be uploaded.")
        # API Gateway conf
        init.add_argument("-api", "--api-gateway-name",
                          help="API Gateway name created to launch the lambda function")

    def _add_invoke_parser(self):
        invoke = self.subparser.add_parser('invoke',
                                           parents=self._get_parents(INVOKE_PARENTS),
                                           help="Call a lambda function using an HTTP request")
        # Set default function
        invoke.set_defaults(func='invoke')
        group = invoke.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name",
                           help="Lambda function name")
        group.add_argument("-f", "--conf-file",
                           help="Yaml file with the function configuration")
        invoke.add_argument("-db", "--data-binary",
                            help="File path of the HTTP data to POST.")
        invoke.add_argument("-jd", "--json-data",
                            help="JSON Body to Post")
        invoke.add_argument("-p", "--parameters",
                            help=("In addition to passing the parameters in the URL, "
                                  "you can pass the parameters here (i.e. '{\"key1\": "
                                  "\"value1\", \"key2\": [\"value2\", \"value3\"]}')."))

    def _add_run_parser(self):
        run = self.subparser.add_parser('run',
                                        parents=self._get_parents(RUN_PARENTS),
                                        help="Deploy function")
        # Set default function
        run.set_defaults(func='run')
        group = run.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name", help="Lambda function name")
        group.add_argument("-f", "--conf-file", help="Yaml file with the function configuration")
        run.add_argument("-s", "--run-script", help="Path to the script passed to the function")
        run.add_argument("-ib", "--input-bucket", help=("Bucket name with files to launch the function."))
        run.add_argument('c_args',
                         nargs=argparse.REMAINDER,
                         help="Arguments passed to the container.")

    def _add_rm_parser(self):
        rm = self.subparser.add_parser('rm',
                                       parents=self._get_parents(RM_LS_PARENTS),
                                       help="Delete function")
        # Set default function
        rm.set_defaults(func='rm')
        group = rm.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name",
                           help="Lambda function name")
        group.add_argument("-a", "--all",
                           help="Delete all lambda functions",
                           action="store_true")
        group.add_argument("-f", "--conf-file",
                           help="Yaml file with the function configuration")

    def _add_log_parser(self):
        log = self.subparser.add_parser('log',
                                        parents=self._get_parents(LOG_PARENTS),
                                        help="Show the logs for the lambda function")
        # Set default function
        log.set_defaults(func='log')
        group = log.add_mutually_exclusive_group(required=True)
        group.add_argument("-n", "--name",
                           help="Lambda function name")
        group.add_argument("-f", "--conf-file",
                           help="Yaml file with the function configuration")
        # CloudWatch args
        log.add_argument("-ls", "--log-stream-name",
                         help="Return the output for the log stream specified.")
        log.add_argument("-ri", "--request-id",
                         help="Return the output for the request id specified.")

    def _add_ls_parser(self):
        ls = self.subparser.add_parser('ls',
                                       parents=self._get_parents(RM_LS_PARENTS),
                                       help="List lambda functions")
        # Set default function
        ls.set_defaults(func='ls')
        # S3 args
        ls.add_argument("-b", "--bucket", help="Show bucket files")

    def _add_put_parser(self):
        put = self.subparser.add_parser('put',
                                        parents=self._get_parents(PUT_GET_PARENTS),
                                        help="Upload file(s) to bucket")
        # Set default function
        put.set_defaults(func='put')

    def _add_get_parser(self):
        get = self.subparser.add_parser('get',
                                        parents=self._get_parents(PUT_GET_PARENTS),
                                        help="Download file(s) from bucket")
        # Set default function
        get.set_defaults(func='get')




# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Module containing all the custom exceptions for SCAR. """

import functools
import sys
from botocore.exceptions import ClientError


def exception(logger):
    '''
    A decorator that wraps the passed in function and logs exceptions
    @param logger: The logging object
    '''

    def decorator(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)

            except ClientError as cerr:
                print(f"There was an exception in {func.__name__}")
                print(cerr.response['Error']['Message'])
                logger.exception(cerr)
                sys.exit(1)

            except ScarError as serr:
                print(serr.args[0])
                logger.exception(serr)
                # Finish the execution if it's an error
                if 'Error' in serr.__class__.__name__:
                    sys.exit(1)

        return wrapper

    return decorator


class ScarError(Exception):
    """
    The base exception class for ScarError exceptions.

    :ivar msg: The descriptive message associated with the error.
    """
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


################################################
##             GENERAL EXCEPTIONS             ##
################################################
class MissingCommandError(ScarError):
    """
    SCAR was launched without a command

    """
    fmt = ("Please use one of the scar available commands "
           "(init,invoke,run,update,rm,ls,log,put,get)")


class ScarConfigFileError(ScarError):
    """
    The SCAR configuration file does not exist and it has been created

    :ivar file_path: Path of the file
    """
    fmt = ("Config file '{file_path}' created.\n"
           "Please, set a valid iam role in the file field 'role' before the first execution.")


class UploadFileNotFoundError(ScarError):
    """
    The file does not exist

    :ivar file_path: Path of the file
    """
    fmt = "Unable to find the file to upload with path '{file_path}'"


class YamlFileNotFoundError(ScarError):
    """
    The yaml configuration file does not exist

    :ivar file_path: Path of the file
    """
    fmt = "Unable to find the yaml file '{file_path}'"


class FdlFileNotFoundError(ScarError):
    """
    The configuration file does not exist

    :ivar file_path: Path of the file
    """
    fmt = "Unable to find the configuration file '{file_path}'"


class ValidatorError(ScarError):
    """
    An error occurred when validating a parameter

    :ivar parameter: Name of the parameter evaluated
    :ivar parameter_value: Current value of the validated parameter
    :ivar error_msg: General error message
    """
    fmt = ("Error validating '{parameter}'.\n"
           "Value '{parameter_value}' incorrect.\n"
           "{error_msg}")


class ScarFunctionNotFoundError(ScarError):
    """
    The called function was not found

    :ivar func_name: Name of the function called
    """
    fmt = "Unable to find the function '{func_name}'"


class FunctionCodeSizeError(ScarError):
    """
    Function code size exceeds AWS limits

    :ivar code_size: Name of the parameter evaluated
    """
    fmt = ("Payload size greater than {code_size}.\n"
           "Please reduce the payload size or use an S3 bucket and try again.")


class S3CodeSizeError(ScarError):
    """
    Function code uploaded to S3 exceeds AWS limits

    :ivar code_size: Name of the parameter evaluated
    """

    fmt = ("Uncompressed image size greater than {code_size}.\n"
           "Please reduce the uncompressed image and try again.")


class GitHubTagNotFoundError(ScarError):
    """
    The specified tag was not found in the GitHub repository

    :ivar version: Tag used for the search
    """

    fmt = "The tag '{tag}' was not found in the GitHub repository."


class StorageProviderNotSupportedError(ScarError):
    """
    The storage provider parsed is not supported

    :ivar provider: Provider specified
    """
    fmt = "The storage provider '{provider}' is not supported."


class AuthenticationVariableNotSupportedError(ScarError):
    """
    The authentication variable parsed is not supported

    :ivar auth_var: Authentication variable specified
    """
    fmt = "The authentication variable '{auth_var}' is not supported."

################################################
##             LAMBDA EXCEPTIONS              ##
################################################
class FunctionCreationError(ScarError):
    """
    An error occurred when creating the lambda function.

    :ivar function_name: Name of the function
    :ivar error_msg: General error message
    """
    fmt = "Unable to create the function '{function_name}' : {error_msg}"


class FunctionNotFoundError(ScarError):
    """
    The requested function does not exist.

    :ivar function_name: Name of the function
    """
    fmt = "Unable to find the function '{function_name}'"


class FunctionExistsError(ScarError):
    """
    The requested function exists.

    :ivar function_name: Name of the function
    """
    fmt = "Function '{function_name}' already exists"


################################################
##               S3 EXCEPTIONS                ##
################################################
class BucketNotFoundError(ScarError):
    """
    The requested bucket does not exist.

    :ivar bucket_name: Name of the bucket
    """
    fmt = "Unable to find the bucket '{bucket_name}'."


class ExistentBucketWarning(ScarError):
    """
    The bucket already exists

    :ivar bucket_name: Name of the bucket
    """
    fmt = "Using existent bucket '{bucket_name}'."


################################################
##         CLOUDWATCH LOGS EXCEPTIONS         ##
################################################
class ExistentLogGroupWarning(ScarError):
    """
    The requested log group already exists

    :ivar log_group_name: Name of the log group
    """
    fmt = "Using existent log group '{logGroupName}'."


class NotExistentLogGroupWarning(ScarError):
    """
    The requested log group does not exists

    :ivar log_group_name: Name of the log group
    """
    fmt = "The requested log group '{logGroupName}' does not exist."


################################################
##           API GATEWAY EXCEPTIONS           ##
################################################
class ApiEndpointNotFoundError(ScarError):
    """
    The requested function does not have an associated API.

    :ivar function_name: Name of the function
    """
    fmt = ("Error retrieving API ID for lambda function '{function_name}'\n"
           "Looks like he requested function does not have an associated API.")


class InvocationPayloadError(ScarError):
    """
    Error invocating the API endpoint.

    :ivar file_size: Size of the passed file
    :ivar max_size: Max size allowd of the file
    """
    fmt = ("Invalid request: Payload size {file_size} greater than {max_size}\n"
           "Check AWS Lambda invocation limits in : "
           "https://docs.aws.amazon.com/lambda/latest/dg/limits.html")


class NotExistentApiGatewayWarning(ScarError):
    """
    The API with the id 'restApiId' was not found.

    :ivar bucket_name: Name of the bucket
    """
    fmt = "The requested API '{restApiId}' does not exist."


################################################
##               IAM EXCEPTIONS               ##
################################################
class GetUserInfoError(ScarError):
    """
    There was an error gettting the IAM user info

    :ivar error_msg: General error message
    """
    fmt = "Error getting the AWS user information.\n{error_msg}."


################################################
##              BATCH EXCEPTIONS              ##
################################################
class InvalidComputeEnvironmentError(ScarError):
    """
    There was an error creating the Batch Compute Environment

    :ivar error_msg: General error message
    """
    fmt = "Error creating the AWS Batch Compute Environment\n."


################################################
##              OSCAR EXCEPTIONS              ##
################################################
class ServiceCreationError(ScarError):
    """
    There was an error creating the OSCAR service

    :ivar service_name: Name of the function
    :ivar error_msg: General error message
    """
    fmt = "Unable to create the service '{service_name}': {error_msg}"


class ServiceDeletionError(ScarError):
    """
    There was an error deleting the OSCAR service

    :ivar service_name: Name of the function
    :ivar error_msg: General error message
    """
    fmt = "Unable to delete the service '{service_name}': {error_msg}"

class ServiceNotFoundError(ScarError):
    """
    There was an error getting the OSCAR service

    :ivar service_name: Name of the function
    :ivar error_msg: General error message
    """
    fmt = "The service '{service_name}' does not exist: {error_msg}"

class ListServicesError(ScarError):
    """
    There was an error getting the OSCAR service

    :ivar service_name: Name of the function
    :ivar error_msg: General error message
    """
    fmt = "Unable to list services from OSCAR cluster '{cluster_id}': {error_msg}"


# author: Asmaa Mirkhan ~ 2019

import os
import argparse
import cv2 as cv
from DetectorAPI import DetectorAPI

def blurBoxes(image, boxes):
    """
    Argument:
    image -- the image that will be edited as a matrix
    boxes -- list of boxes that will be blurred, each box must be int the format (x_top_left, y_top_left, x_bottom_right, y_bottom_right)
    
    Returns:
    image -- the blurred image as a matrix
    """

    for box in boxes:
        # unpack each box
        x1, y1, x2, y2 = [d for d in box]

        # crop the image due to the current box
        sub = image[y1:y2, x1:x2]

        # apply GaussianBlur on cropped area
        blur = cv.blur(sub, (10, 10))

        # paste blurred image on the original image
        image[y1:y2, x1:x2] = blur

    return image


def main(args):
    # assign model path and threshold
    model_path = args.model_path
    threshold = args.threshold

    # create detection object
    odapi = DetectorAPI(path_to_ckpt=model_path)

    # open image
    image = cv.imread(args.input_image)

    # real face detection
    boxes, scores, classes, num = odapi.processFrame(image)

    # filter boxes due to threshold
    # boxes are in (x_top_left, y_top_left, x_bottom_right, y_bottom_right) format
    boxes = [boxes[i] for i in range(0, num) if scores[i] > threshold]

    # apply blurring
    image = blurBoxes(image, boxes)

    # # show image
    # cv.imshow('blurred', image)

    # if image will be saved then save it
    if args.output_image:
        cv.imwrite(args.output_image, image)
        print('Image has been saved successfully at', args.output_image,
              'path')
    else:
        cv.imshow('blurred', image)
        # when any key has been pressed then close window and stop the program
        cv.waitKey(0)
        cv.destroyAllWindows()


if __name__ == "__main__":
    # creating argument parser
    parser = argparse.ArgumentParser(description='Image blurring parameters')

    # adding arguments
    parser.add_argument('-i',
                        '--input_image',
                        help='Path to your image',
                        type=str,
                        required=True)
    parser.add_argument('-m',
                        '--model_path',
                        default='/opt/blurry-faces/face_model/face.pb',
                        help='Path to .pb model',
                        type=str)
    parser.add_argument('-o',
                        '--output_image',
                        help='Output file path',
                        type=str)
    parser.add_argument('-t',
                        '--threshold',
                        help='Face detection confidence',
                        default=0.7,
                        type=float)
    args = parser.parse_args()
    print(args)
    # if input image path is invalid then stop
    assert os.path.isfile(args.input_image), 'Invalid input file'

    # if output directory is invalid then stop
    if args.output_image:
        assert os.path.isdir(os.path.dirname(
            args.output_image)), 'No such directory'

    main(args)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class implementing the low-level functions to 
communicate with an OSCAR cluster."""

from typing import Dict, List
import scar.logger as logger
import scar.exceptions as excp
import requests

# Disable InsecureRequestWarning in requests package
requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)


def _get_error_msg(res: requests.Response) -> str:
    error_msg = ''
    if res.text:
        error_msg = res.text
    elif res.status_code == 400:
        error_msg = 'Bad Request'
    elif res.status_code == 401:
        error_msg = 'Invalid Credentials'
    elif res.status_code == 404:
        error_msg = 'The Service doesn\'t exist'
    elif res.status_code == 500:
        error_msg = 'Internal Server Error'
    return error_msg    


class OSCARClient():
    _SERVICES_PATH = '/system/services'

    def __init__(self, credentials_info: Dict, cluster_id: str):
        self.cluster_id = cluster_id
        self.endpoint = credentials_info['endpoint']
        self.auth_user = credentials_info['auth_user']
        self.auth_password = credentials_info['auth_password']
        self.ssl_verify = credentials_info['ssl_verify']
        
    def create_service(self, **kwargs: Dict) -> Dict:
        """Creates a new OSCAR service."""
        logger.debug('Creating OSCAR service.')
        res = requests.post(
            f'{self.endpoint}{self._SERVICES_PATH}',
            auth=(self.auth_user, self.auth_password),
            verify=self.ssl_verify,
            json=kwargs
        )
        # Raise a ServiceCreationError if the return code is not 201
        if res.status_code != 201:
            raise excp.ServiceCreationError(service_name=kwargs['name'], error_msg=_get_error_msg(res))

    def delete_service(self, service_name: str) -> None:
        """Deletes an OSCAR service."""
        logger.debug('Deleting OSCAR service.')
        res = requests.delete(
            f'{self.endpoint}{self._SERVICES_PATH}/{service_name}',
            auth=(self.auth_user, self.auth_password),
            verify=self.ssl_verify
        )
        # Raise a ServiceDeletionError if the return code is not 204
        if res.status_code != 204:
            raise excp.ServiceDeletionError(service_name=service_name, error_msg=_get_error_msg(res))

    def get_service(self, service_name: str) -> Dict:
        """Get the properties of the specified service."""
        res = requests.get(
            f'{self.endpoint}{self._SERVICES_PATH}/{service_name}',
            auth=(self.auth_user, self.auth_password),
            verify=self.ssl_verify
        )
        # Raise a ServiceNotFoundError if the return code is not 200
        if res.status_code != 200:
            raise excp.ServiceNotFoundError(service_name=service_name, error_msg=_get_error_msg(res))
        return res.json()

    def list_services(self) -> List:
        """Get all the services registered in the cluster."""
        res = requests.get(
            f'{self.endpoint}{self._SERVICES_PATH}',
            auth=(self.auth_user, self.auth_password),
            verify=self.ssl_verify
        )
        # Raise a ListServicesError if the return code is not 200
        if res.status_code != 200:
            raise excp.ListServicesError(cluster_id=self.cluster_id, error_msg=_get_error_msg(res))
        return res.json()



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
IAM creation, deletion and configuration."""

from typing import Dict
from botocore.exceptions import ClientError
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception, GetUserInfoError
import scar.logger as logger
from scar.utils import StrUtils


class IAMClient(BotoClient):
    """A low-level client representing aws Identity and Access Management (IAMClient).
    DOC_URL: https://boto3.readthedocs.io/en/latest/reference/services/iam.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'iam'
    _USER_NAME_REGEX = r'(?<=user\/)(\S+)'

    @exception(logger)
    def get_user_info(self) -> Dict:
        """Retrieves information about the specified IAM user,
        including the user's creation date, path, unique ID, and ARN."""
        try:
            return self.client.get_user()
        except ClientError as cerr:
            if cerr.response['Error']['Code'] == 'AccessDenied':
                # If the user doesn't have access rights to IAMClient
                # we can find the user name in the error response
                user_name = StrUtils.find_expression(str(cerr), self._USER_NAME_REGEX)
                return {'UserName' : user_name,
                        'User' : {'UserName' : user_name,
                                  'UserId' : ''}}
            raise cerr
        except Exception as ex:
            raise GetUserInfoError(error_msg=ex)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with class and methods used to manage the OSCAR provider."""

from typing import Dict
from copy import deepcopy
from scar.utils import StrUtils, FileUtils, SupervisorUtils
from scar.providers.oscar.client import OSCARClient
from scar.providers.aws.controller import add_output
import scar.exceptions as excp
import scar.logger as logger
import scar.providers.oscar.response as response_parser


def _get_creation_args(resources_info: Dict, storage_providers: Dict) -> Dict:
    creation_args = {}
    # Clean None values
    for k,v in resources_info.items():
        if v:
            creation_args[k] = v
    # Get the content of 'script'
    creation_args['script'] = FileUtils.read_file(creation_args['script'])
    # Add storage_providers
    creation_args['storage_providers'] = storage_providers
    return creation_args


def _get_credentials_info(resources_info: Dict) -> Dict:
    return {
        'endpoint': resources_info.get('endpoint', ''),
        'auth_user': resources_info.get('auth_user', ''),
        'auth_password': resources_info.get('auth_password', ''),
        'ssl_verify': resources_info.get('ssl_verify', True)
    }

def _are_credentials_defined(credentials_info: Dict) -> bool:
    return (credentials_info['endpoint'] and
            credentials_info['auth_user'] and
            credentials_info['auth_password'])


class OSCAR():

    def __init__(self, func_call: Dict):
        self.raw_args = FileUtils.load_tmp_config_file()
        # Flatten the list of services
        self.oscar_resources = []
        nested_resources = self.raw_args.get('functions', {}).get('oscar', [])
        for resources in nested_resources:
            for cluster_id, resources_info in resources.items():
                if ('name' in resources[cluster_id] and
                        resources[cluster_id]['name']):
                    resources_info['cluster_id'] = cluster_id
                    self.oscar_resources.append(resources_info)
        # Store the storage_providers dict independently
        self.storage_providers = self.raw_args.get('storage_providers', {})
        self.scar_info = self.raw_args.get('scar', {})
        add_output(self.scar_info)
        # Call the user's command
        getattr(self, func_call)()

    @excp.exception(logger)
    def init(self):
        for resources_info in self.oscar_resources:
            resources_info = deepcopy(resources_info)
            self._create_oscar_service(resources_info)
            response_parser.parse_service_creation(resources_info, self.scar_info.get('cli_output'))

    @excp.exception(logger)
    def rm(self):
        for resources_info in self.oscar_resources:
            resources_info = deepcopy(resources_info)
            credentials_info = _get_credentials_info(resources_info)
            OSCARClient(credentials_info, resources_info.get('cluster_id', '')).delete_service(resources_info['name'])
            response_parser.parse_service_deletion(resources_info, self.scar_info.get('cli_output'))

    def _create_oscar_service(self, resources_info: Dict):
        credentials_info = _get_credentials_info(resources_info)
        creation_args = _get_creation_args(resources_info, self.storage_providers)
        OSCARClient(credentials_info, resources_info.get('cluster_id', '')).create_service(**creation_args)

    @excp.exception(logger)
    def ls(self):
        clusters = self.raw_args.get('functions', {}).get('oscar', [{}])[0]
        for cluster_id, resources_info in clusters.items():
            credentials_info = _get_credentials_info(resources_info)
            if _are_credentials_defined(credentials_info):
                oscar_resources = OSCARClient(credentials_info, cluster_id).list_services()
                response_parser.parse_ls_response(oscar_resources,
                                                  credentials_info['endpoint'],
                                                  cluster_id,
                                                  self.scar_info.get('cli_output'))



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with classes and methods to manage the
CloudWatch Log functionalities at high level."""

from typing import List, Dict
from botocore.exceptions import ClientError
from scar.providers.aws import GenericClient
from scar.providers.aws.batchfunction import Batch
import scar.logger as logger


def _parse_events_in_message(log_events: List) -> str:
    data = [(event.get('message', ''), event.get('timestamp', '')) for event in log_events]
    sorted_data = sorted(data, key=lambda time: time[1])
    return "".join([sdata[0] for sdata in sorted_data])


class CloudWatchLogs(GenericClient):
    """Manages the AWS CloudWatch Logs functionality"""
    
    def __init__(self, resources_info: Dict):
        super().__init__(resources_info.get('cloudwatch'))
        self.resources_info = resources_info
        self.cloudwatch = resources_info.get('cloudwatch')

    def get_log_group_name(self, function_name: str=None) -> str:
        """Returns the log group matching the
        current lambda function being parsed."""
        if function_name:
            return f'/aws/lambda/{function_name}'
        return f'/aws/lambda/{self.resources_info.get("lambda").get("name")}'

    def _get_log_group_name_arg(self, function_name: str=None) -> Dict:
        return {'logGroupName' : self.get_log_group_name(function_name)}

    def _is_end_line(self, line: str) -> bool:
        return line.startswith('REPORT') and self.cloudwatch.get('request_id') in line

    def _is_start_line(self, line: str) -> bool:
        return line.startswith('START') and self.cloudwatch.get('request_id') in line

    def _parse_logs_with_requestid(self, function_logs: str) -> str:
        parsed_msg = ""
        if function_logs:
            in_req_id_logs = False
            for line in function_logs.split('\n'):
                if self._is_start_line(line):
                    parsed_msg += f'{line}\n'
                    in_req_id_logs = True
                elif self._is_end_line(line):
                    parsed_msg += line
                    break
                elif in_req_id_logs:
                    parsed_msg += f'{line}\n'
        return parsed_msg

    def _get_lambda_logs(self):
        """Returns Lambda logs for an specific lambda function."""
        function_logs = ""
        try:
            kwargs = self._get_log_group_name_arg()
            if self.cloudwatch.get("log_stream_name", False):
                kwargs["logStreamNames"] = [self.cloudwatch.get("log_stream_name")]
            function_logs = _parse_events_in_message(self.client.get_log_events(**kwargs))
            if self.cloudwatch.get("request_id", False):
                function_logs = self._parse_logs_with_requestid(function_logs)
        except ClientError as cerr:
            logger.warning("Error getting the function logs: %s" % cerr)
        return function_logs
            
    def _get_batch_job_log(self, jobs_info: List) -> str:
        """Returns Batch logs for an specific job."""
        batch_logs = ""
        if jobs_info:
            job = jobs_info[0]
            batch_logs += f"Batch job status: {job.get('status', '')}\n"
            kwargs = {'logGroupName': "/aws/batch/job"}
            if job.get("status", "") == "SUCCEEDED":
                kwargs['logStreamNames'] = [job.get("container", {}).get("logStreamName", "")]
                batch_events = self.client.get_log_events(**kwargs)
                msgs = [event.get('message', '')
                        for response in batch_events
                        for event in response.get("events", {})]
                batch_logs += '\n'.join(msgs)
        return batch_logs

    def create_log_group(self) -> Dict:
        """Creates a CloudWatch Log Group."""
        creation_args = self._get_log_group_name_arg()
        creation_args['tags'] = self.resources_info.get('lambda').get('tags')
        response = self.client.create_log_group(**creation_args)
        # Set retention policy into the log group
        retention_args = self._get_log_group_name_arg()
        retention_args['retentionInDays'] = self.cloudwatch.get('log_retention_policy_in_days')
        self.client.set_log_retention_policy(**retention_args)
        return response

    def delete_log_group(self, log_group_name: str) -> Dict:
        """Deletes a CloudWatch Log Group."""
        return self.client.delete_log_group(log_group_name)

    def get_aws_logs(self) -> str:
        """Returns Cloudwatch logs for an specific lambda function and batch job (if any)."""
        aws_logs = self._get_lambda_logs()
        batch_logs = ""
        if self.resources_info.get('cloudwatch').get('request_id', False):
            batch_jobs = Batch(self.resources_info).get_jobs_with_request_id()
            batch_logs = self._get_batch_job_log(batch_jobs["jobs"])
        return aws_logs + batch_logs if batch_logs else aws_logs



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module with the class necessary to manage the
EC2 creation, deletion and configuration."""

from typing import Dict
from scar.providers.aws.clients import BotoClient
from scar.exceptions import exception
import scar.logger as logger


class EC2Client(BotoClient):
    """A low-level client representing Amazon Elastic Compute Cloud (EC2).
    DOC_URL: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html"""

    # Parameter used by the parent to create the appropriate boto3 client
    _BOTO_CLIENT_NAME = 'ec2'

    @exception(logger)
    def create_launch_template(self, name: str, description: str, data: Dict) -> Dict:
        """Creates a launch template.
        A launch template contains the parameters to launch an instance."""
        kwargs = {'LaunchTemplateName': name,
                  'VersionDescription': description,
                  'LaunchTemplateData': data}
        return self.client.create_launch_template(**kwargs)

    @exception(logger)
    def create_launch_template_version(self, name: str, description: str, data: Dict) -> Dict:
        """Creates a new version for a launch template.
        You can specify an existing version of launch template
        from which to base the new version."""
        kwargs = {'LaunchTemplateName': name,
                  'VersionDescription': description,
                  'LaunchTemplateData': data}
        return self.client.create_launch_template_version(**kwargs)

    @exception(logger)
    def describe_launch_templates(self, parameters: Dict) -> Dict:
        """Describes one or more launch templates."""
        return self.client.describe_launch_templates(**parameters)

    @exception(logger)
    def describe_launch_template_versions(self, parameters: Dict) -> Dict:
        """Describes one or more versions of a specified launch template."""
        return self.client.describe_launch_template_versions(**parameters)
